from .api.openapi_generated import (
    TableauError,
    DataType,
    Connection,
    Datasource,
    FilterType,
    DimensionFilterField,
    CalculatedFilterField,
    Function,
    QuantitativeFilterType,
    QueryOutput,
    ReturnFormat,
    SortDirection,
    PeriodType,
    DateRangeType,
    Direction,
    FieldBase,
    DimensionField,
    MeasureField,
    CalculatedField,
    FieldMetadata,
    MeasureFilterField,
    MetadataOutput,
    QueryOptions,
    ReadMetadataRequest,
    Field,
    FilterField,
    QueryDatasourceOptions,
    Filter,
    MatchFilter,
    QuantitativeFilterBase,
    QuantitativeNumericalFilter,
    QuantitativeDateFilter,
    Query,
    QueryRequest,
    SetFilter,
    RelativeDateFilter,
    TopNFilter,
)
from .api import (
    VizQLDataServiceClient,
    read_metadata,
    query_datasource,
)

__all__ = [
    # API related
    "VizQLDataServiceClient",
    "read_metadata",
    "query_datasource",
    # OpenAPI models
    "TableauError",
    "DataType",
    "Connection",
    "Datasource",
    "FilterType",
    "DimensionFilterField",
    "CalculatedFilterField",
    "Function",
    "QuantitativeFilterType",
    "QueryOutput",
    "ReturnFormat",
    "SortDirection",
    "PeriodType",
    "DateRangeType",
    "Direction",
    "FieldBase",
    "DimensionField",
    "MeasureField",
    "CalculatedField",
    "FieldMetadata",
    "MeasureFilterField",
    "MetadataOutput",
    "QueryOptions",
    "ReadMetadataRequest",
    "Field",
    "FilterField",
    "QueryDatasourceOptions",
    "Filter",
    "MatchFilter",
    "QuantitativeFilterBase",
    "QuantitativeNumericalFilter",
    "QuantitativeDateFilter",
    "Query",
    "QueryRequest",
    "SetFilter",
    "RelativeDateFilter",
    "TopNFilter",
]



"""
Utility Module

This module provides utility functions for the SDK.
"""


def format_server_url(url: str, subdomain: str = "") -> str:
    """Format server URL with proper protocol and subdomain.

    Args:
        url: The server URL to format
        subdomain: Optional subdomain to append to the URL

    Returns:
        str: Formatted URL with proper protocol and subdomain

    Raises:
        ValueError: If URL is empty or contains spaces
        TypeError: If URL is None
    """
    if url is None:
        raise TypeError("URL must be a string")

    if not url:
        raise ValueError("URL cannot be empty")

    if " " in url:
        raise ValueError("Invalid URL format")

    # Handle server URL protocol and domain
    if "online.tableau.com" in url and not url.startswith("https"):
        url = f"https://{url}"
    elif not url.startswith(("http://", "https://")):
        url = f"http://{url}"

    # Remove trailing slash if present
    url = url.rstrip("/")

    # Append subdomain if provided
    if subdomain:
        # Ensure subdomain starts with a slash
        if not subdomain.startswith("/"):
            subdomain = f"/{subdomain}"
        url = f"{url}{subdomain}"

    return url



from http import HTTPStatus
from typing import Any, Optional

import httpx

from .client import VizQLDataServiceClient
from .errors import UnexpectedStatus
from .openapi_generated import MetadataOutput, ReadMetadataRequest
from .types import Response


def _get_kwargs(
    *,
    body: ReadMetadataRequest,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/read-metadata",
    }

    _body = body.model_dump(mode="json", exclude_none=True)
    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: VizQLDataServiceClient, response: httpx.Response
) -> Optional[MetadataOutput]:
    if response.status_code == 200:
        response_200 = MetadataOutput.model_validate_json(response.content)

        return response_200
    if client.raise_on_unexpected_status:
        raise UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: VizQLDataServiceClient, response: httpx.Response
) -> Response[MetadataOutput]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: VizQLDataServiceClient,
    body: ReadMetadataRequest,
) -> Response[MetadataOutput]:
    """Request data source metadata with detailed response information

     Requests metadata for a specific data source and returns a detailed response containing:
     - The metadata information (MetadataOutput)
     - HTTP status code
     - Response headers
     - Raw response content

     The metadata provides information about the data fields, such as field names, data types, and descriptions.

    Args:
        body (ReadMetadataRequest): The metadata request parameters

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[MetadataOutput]: A response object containing both the metadata and response metadata
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: VizQLDataServiceClient,
    body: ReadMetadataRequest,
) -> Optional[MetadataOutput]:
    """Request data source metadata and get only the metadata information

     Requests metadata for a specific data source and returns only the metadata information without response metadata.
     This is a convenience wrapper around sync_detailed() that returns only the parsed metadata.

     The metadata provides information about the data fields, such as field names, data types, and descriptions.

    Args:
        body (ReadMetadataRequest): The metadata request parameters

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Optional[MetadataOutput]: The metadata information, or None if the request was unsuccessful
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: VizQLDataServiceClient,
    body: ReadMetadataRequest,
) -> Response[MetadataOutput]:
    """Request data source metadata asynchronously with detailed response information

     Asynchronously requests metadata for a specific data source and returns a detailed response containing:
     - The metadata information (MetadataOutput)
     - HTTP status code
     - Response headers
     - Raw response content

     The metadata provides information about the data fields, such as field names, data types, and descriptions.

    Args:
        body (ReadMetadataRequest): The metadata request parameters

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[MetadataOutput]: A response object containing both the metadata and response metadata
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: VizQLDataServiceClient,
    body: ReadMetadataRequest,
) -> Optional[MetadataOutput]:
    """Request data source metadata asynchronously and get only the metadata information

     Asynchronously requests metadata for a specific data source and returns only the metadata information without response metadata.
     This is a convenience wrapper around asyncio_detailed() that returns only the parsed metadata.

     The metadata provides information about the data fields, such as field names, data types, and descriptions.

    Args:
        body (ReadMetadataRequest): The metadata request parameters

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Optional[MetadataOutput]: The metadata information, or None if the request was unsuccessful
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed


__all__ = [
    "sync",
    "sync_detailed",
    "asyncio",
    "asyncio_detailed",
]



import argparse
import inspect
import os
import sys
import traceback
from typing import Optional, Union

import tableauserverclient as TSC

# Add project root to sys.path
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, root_dir)

# Determine if we're in development or production environment
is_development = os.path.basename(root_dir) == "python_sdk"

if is_development:
    from src.api.openapi_generated import Datasource
    from src.api.utils import format_server_url
else:
    from vizql_data_service_py.api.openapi_generated import Datasource  # type: ignore
    from vizql_data_service_py.api.utils import format_server_url  # type: ignore

SAMPLE_DATASOURCE = "Superstore Datasource"


def print_help():
    """Print help information for all available commands."""
    print("\n=== VizQL Data Service Python SDK Examples ===")
    print("\nBasic Usage:")
    print("  python examples.py [options]")
    print("  python examples.py --async [options]  # Run in async mode")

    print("\nRequired Arguments:")
    print("  -s, --server URL    Tableau Server URL (required)")

    print("\nAuthentication Options (choose one authentication):")
    print("  -u, --user USERNAME        Tableau Server username")
    print("  -p, --password PASSWORD    Tableau Server password")
    print("  -n, --pat-name NAME        Personal Access Token name")
    print("  -t, --pat-secret SECRET    Personal Access Token secret")
    print("  -j, --jwt-token TOKEN      JWT token")

    print("\nOptional Arguments:")
    print(
        "  -S, --site SITE_NAME       Tableau Server site name, or the default site if unspecified"
    )
    print("  -v, --verbose              Print detailed request response information")
    print("  -h, --help                 Show this help message")

    print("\nExamples:")
    print("  1. Basic usage with username/password:")
    print("     python examples.py -s https://your-server -u admin -p password")
    print("\n  2. Using Personal Access Token:")
    print(
        "     python examples.py -s https://your-server -n token-name -t token-secret"
    )
    print("\n  3. Using JWT token:")
    print("     python examples.py -s https://your-server -j your-jwt-token")
    print("\n  4. Running in async mode with verbose output:")
    print(
        "     python examples.py --async -v -s https://your-server -u admin -p password"
    )


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="VizQL Data Service Python SDK Examples",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,  # Disable default help
    )
    parser.add_argument("-s", "--server", required=True, help="Tableau Server URL")
    parser.add_argument("-u", "--user", help="Tableau Server username")
    parser.add_argument("-p", "--password", help="Tableau Server password")
    parser.add_argument("-n", "--pat-name", help="Personal Access Token name")
    parser.add_argument("-t", "--pat-secret", help="Personal Access Token secret")
    parser.add_argument("-j", "--jwt-token", help="JWT token")
    parser.add_argument("-S", "--site", help="Tableau Server site name")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print detailed information"
    )
    parser.add_argument(
        "-h", "--help", action="store_true", help="Show this help message"
    )
    return parser


def list_datasources_and_get_luid(server: TSC.Server, verbose: bool = False):
    if not server.auth_token:
        raise RuntimeError("Not signed in. Please call sign_in() first.")
    all_datasources, pagination_item = server.datasources.get()

    if verbose:
        print(f"\nThere are {pagination_item.total_available} datasources on site:")
        for ds in all_datasources:
            print(f"- {ds.name} (LUID: {ds.id})")

    matching_datasources = [
        ds for ds in all_datasources if ds.name == SAMPLE_DATASOURCE
    ]
    if not matching_datasources:
        raise ValueError(f"Datasource named '{SAMPLE_DATASOURCE}' not found.")

    selected_ds = matching_datasources[0]
    print(f"\nUsing '{selected_ds.name}' with ID: {selected_ds.id}")
    return selected_ds.id


def create_datasource(luid: str) -> Datasource:
    """Create a Datasource object with the given LUID."""
    return Datasource(datasourceLuid=luid)


def handle_response(response, query_name, verbose=False):
    """Handle the response from the API."""
    print(f"\n{query_name} Response:")

    if verbose:
        print(f"Response Status Code: {response.status_code}")
        print(f"Response Headers: {response.headers}")

    print(f"Response Body: {response.content}")


def handle_error(e, operation_name: str = "Operation", verbose=False):
    """Handle exceptions and print detailed information.

    Args:
        e: The exception
        operation_name: Name of the operation for logging
        verbose: Whether to print detailed error information
    """
    print(f"\n{operation_name} failed: {str(e)}")
    if verbose:
        print_exception_details(e)


def print_exception_details(e):
    """Print detailed exception information"""
    print("\n=== Exception Details ===")
    print(f"Exception Type: {type(e).__name__}")
    print(f"Exception Message: {str(e)}")
    print("\n=== Full Traceback ===")
    traceback.print_exc()
    frame = inspect.currentframe()
    if frame:
        print("\n=== Current Frame Info ===")
        print(f"File: {frame.f_code.co_filename}")
        print(f"Line: {frame.f_lineno}")
        print(f"Function: {frame.f_code.co_name}")
        print("\nLocal Variables:")
        for name, value in frame.f_locals.items():
            print(f"  {name}: {type(value)} = {value}")


def create_server(
    url: str,
    *,
    username: Optional[str] = None,
    password: Optional[str] = None,
    pat_name: Optional[str] = None,
    pat_secret: Optional[str] = None,
    jwt_token: Optional[str] = None,
    site_id: str = "",
) -> tuple[str, Union[TSC.JWTAuth, TSC.PersonalAccessTokenAuth, TSC.TableauAuth]]:
    """Create a server configuration with authentication.

    Args:
        url: The base URL of the server.
        username: Username for Tableau authentication.
        password: Password for Tableau authentication.
        pat_name: Personal Access Token name (--pat-name).
        pat_secret: Personal Access Token secret (--pat-secret).
        jwt_token: JWT token for JWT authentication (--jwt-token).
        site_id: The site ID to use.

    Returns:
        tuple[str, Union[TSC.JWTAuth, TSC.PersonalAccessTokenAuth, TSC.TableauAuth]]:
            A tuple containing the formatted server URL and the authentication object.

    Raises:
        ValueError: If no valid authentication method is provided.
    """
    formatted_url = format_server_url(url)

    # Check for JWT authentication
    if jwt_token:
        auth: Union[TSC.JWTAuth, TSC.PersonalAccessTokenAuth, TSC.TableauAuth] = (
            TSC.JWTAuth(jwt_token, site_id)
        )
    # Check for Personal Access Token authentication
    elif pat_name and pat_secret:
        auth = TSC.PersonalAccessTokenAuth(pat_name, pat_secret, site_id)
    # Check for username/password authentication
    elif username and password:
        auth = TSC.TableauAuth(username, password, site_id)
    else:
        raise ValueError(
            "No valid authentication method provided. Please provide either:\n"
            "- JWT token (--jwt-token)\n"
            "- PAT name and secret (--pat-name, --pat-secret)\n"
            "- Username and password (--user, --password)"
        )

    return formatted_url, auth



import os
import sys
from datetime import date

# Add project root to sys.path
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, root_dir)

# Determine if we're in development or production environment
is_development = os.path.basename(root_dir) == "python_sdk"

if is_development:
    from src.api.openapi_generated import (
        CalculatedField,
        DateRangeType,
        DimensionField,
        DimensionFilterField,
        Direction,
        FilterType,
        Function,
        MatchFilter,
        MeasureField,
        MeasureFilterField,
        PeriodType,
        QuantitativeDateFilter,
        QuantitativeFilterType,
        QuantitativeNumericalFilter,
        Query,
        RelativeDateFilter,
        SetFilter,
        TopNFilter,
    )
else:
    from vizql_data_service_py.api.openapi_generated import (  # type: ignore
        CalculatedField,
        DateRangeType,
        DimensionField,
        DimensionFilterField,
        Direction,
        FilterType,
        Function,
        MatchFilter,
        MeasureField,
        MeasureFilterField,
        PeriodType,
        QuantitativeDateFilter,
        QuantitativeFilterType,
        QuantitativeNumericalFilter,
        Query,
        RelativeDateFilter,
        SetFilter,
        TopNFilter,
    )


def create_simple_query():
    return Query(
        fields=[
            DimensionField(fieldCaption="Category"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ]
    )


def create_custom_calculation():
    return Query(
        fields=[
            CalculatedField(
                fieldCaption="AOV", calculation="SUM([Profit])/COUNTD([Order ID])"
            ),
        ]
    )


def create_dimension_filter():
    return Query(
        fields=[
            DimensionField(fieldCaption="Ship Mode"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            SetFilter(
                field=DimensionFilterField(fieldCaption="Ship Mode"),
                filterType=FilterType.SET,
                values=["First Class", "Standard Class"],
                exclude=False,
            )
        ],
    )


def create_quantitative_range_filter():
    return Query(
        fields=[
            DimensionField(fieldCaption="Ship Mode"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            QuantitativeNumericalFilter(
                field=MeasureFilterField(fieldCaption="Sales", function=Function.SUM),
                filterType=FilterType.QUANTITATIVE_NUMERICAL,
                quantitativeFilterType=QuantitativeFilterType.RANGE,
                min=266839,
                max=1149562,
            )
        ],
    )


def create_quantitative_date_filter():
    return Query(
        fields=[
            MeasureField(fieldCaption="Order Date", function=Function.YEAR),
            MeasureField(fieldCaption="Order Date", function=Function.QUARTER),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            QuantitativeDateFilter(
                field=DimensionFilterField(fieldCaption="Order Date"),
                filterType=FilterType.QUANTITATIVE_DATE,
                quantitativeFilterType=QuantitativeFilterType.MIN,
                minDate=date(2020, 1, 1),
            )
        ],
    )


def create_relative_date_filter():
    return Query(
        fields=[
            MeasureField(
                fieldCaption="Order Date", function=Function.YEAR, sortPriority=1
            ),
            MeasureField(
                fieldCaption="Order Date", function=Function.MONTH, sortPriority=2
            ),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            RelativeDateFilter(
                field=DimensionFilterField(fieldCaption="Order Date"),
                filterType=FilterType.DATE,
                periodType=PeriodType.MONTHS,
                dateRangeType=DateRangeType.CURRENT,
                anchorDate=date(2022, 1, 1),
            )
        ],
    )


def create_match_filter():
    return Query(
        fields=[
            DimensionField(fieldCaption="State/Province"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            MatchFilter(
                field=DimensionFilterField(fieldCaption="State/Province"),
                filterType=FilterType.MATCH,
                startsWith="A",
                endsWith="a",
                contains="o",
                exclude=False,
            )
        ],
    )


def create_top_n_filter():
    return Query(
        fields=[
            DimensionField(fieldCaption="State/Province"),
            MeasureField(fieldCaption="Profit", function=Function.SUM),
        ],
        filters=[
            TopNFilter(
                field=DimensionFilterField(fieldCaption="State/Province"),
                filterType=FilterType.TOP,
                howMany=10,
                fieldToMeasure=MeasureFilterField(
                    fieldCaption="Profit", function=Function.SUM
                ),
                direction=Direction.TOP,
            )
        ],
    )


def create_multiple_dimension_filters():
    return Query(
        fields=[
            DimensionField(fieldCaption="Ship Mode"),
            DimensionField(fieldCaption="Segment"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            SetFilter(
                field=DimensionFilterField(fieldCaption="Ship Mode"),
                filterType=FilterType.SET,
                values=["First Class", "Standard Class"],
                exclude=False,
            ),
            SetFilter(
                field=DimensionFilterField(fieldCaption="Segment"),
                filterType=FilterType.SET,
                values=["Consumer"],
                exclude=True,
            ),
        ],
    )


def create_multiple_min_max_numeric_filters():
    return Query(
        fields=[
            DimensionField(fieldCaption="Ship Mode"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
            MeasureField(fieldCaption="Profit", function=Function.SUM),
        ],
        filters=[
            QuantitativeNumericalFilter(
                field=MeasureFilterField(fieldCaption="Sales", function=Function.SUM),
                filterType=FilterType.QUANTITATIVE_NUMERICAL,
                quantitativeFilterType=QuantitativeFilterType.MIN,
                min=266839,
            ),
            QuantitativeNumericalFilter(
                field=MeasureFilterField(fieldCaption="Profit", function=Function.SUM),
                filterType=FilterType.QUANTITATIVE_NUMERICAL,
                quantitativeFilterType=QuantitativeFilterType.MAX,
                max=164098,
            ),
        ],
    )


def create_dimension_numeric_filters():
    return Query(
        fields=[
            DimensionField(fieldCaption="Ship Mode"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            SetFilter(
                field=DimensionFilterField(fieldCaption="Ship Mode"),
                filterType=FilterType.SET,
                values=["First Class", "Standard Class"],
                exclude=True,
            ),
            QuantitativeNumericalFilter(
                field=MeasureFilterField(fieldCaption="Profit", function=Function.SUM),
                filterType=FilterType.QUANTITATIVE_NUMERICAL,
                quantitativeFilterType=QuantitativeFilterType.MIN,
                min=40000,
            ),
        ],
    )


def create_context_filter():
    return Query(
        fields=[
            DimensionField(fieldCaption="Sub-Category"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
        ],
        filters=[
            TopNFilter(
                field=DimensionFilterField(fieldCaption="Sub-Category"),
                filterType=FilterType.TOP,
                howMany=10,
                fieldToMeasure=MeasureFilterField(
                    fieldCaption="Sales", function=Function.SUM
                ),
                direction=Direction.TOP,
            ),
            SetFilter(
                field=DimensionFilterField(fieldCaption="Category"),
                filterType=FilterType.SET,
                values=["Furniture"],
                exclude=False,
                context=True,
            ),
        ],
    )


def create_numeric_date_dimension_filters():
    return Query(
        fields=[
            DimensionField(fieldCaption="Order Date"),
            MeasureField(fieldCaption="Sales", function=Function.SUM),
            DimensionField(fieldCaption="Ship Mode"),
        ],
        filters=[
            QuantitativeNumericalFilter(
                field=MeasureFilterField(fieldCaption="Sales", function=Function.SUM),
                filterType=FilterType.QUANTITATIVE_NUMERICAL,
                quantitativeFilterType=QuantitativeFilterType.RANGE,
                min=10,
                max=63,
            ),
            RelativeDateFilter(
                field=DimensionFilterField(fieldCaption="Order Date"),
                filterType=FilterType.DATE,
                periodType=PeriodType.MONTHS,
                dateRangeType=DateRangeType.NEXTN,
                rangeN=3,
                anchorDate=date(2022, 1, 1),
            ),
            SetFilter(
                field=DimensionFilterField(fieldCaption="Ship Mode"),
                filterType=FilterType.SET,
                values=["First Class"],
                exclude=False,
            ),
        ],
    )


QUERY_FUNCTIONS = [
    create_simple_query,
    create_custom_calculation,
    create_dimension_filter,
    create_quantitative_range_filter,
    create_quantitative_date_filter,
    create_relative_date_filter,
    create_match_filter,
    create_top_n_filter,
    create_multiple_dimension_filters,
    create_multiple_min_max_numeric_filters,
    create_dimension_numeric_filters,
    create_context_filter,
    create_numeric_date_dimension_filters,
]



import ssl
from importlib.metadata import version
from typing import Any, Optional, Union

import httpx
import tableauserverclient as TSC
from attrs import define, evolve, field

from .utils import format_server_url

API_SUBDOMAIN = "/api/v1/vizql-data-service"
X_TABLEAU_AUTH = "X-Tableau-Auth"
VERSION = version("vizql-data-service-py")  # Read version from package metadata


@define
class AuthenticatedClient:
    """A Client which has been authenticated for use on secured endpoints

    The following are accepted as keyword arguments and will be used to construct httpx Clients internally:

        ``base_url``: The base URL for the API, all requests are made to a relative path to this URL

        ``cookies``: A dictionary of cookies to be sent with every request

        ``headers``: A dictionary of headers to be sent with every request

        ``timeout``: The maximum amount of a time a request can take. API functions will raise
        httpx.TimeoutException if this is exceeded.

        ``verify_ssl``: Whether or not to verify the SSL certificate of the API server. This should be True in production,
        but can be set to False for testing purposes.

        ``follow_redirects``: Whether or not to follow redirects. Default value is False.

        ``httpx_args``: A dictionary of additional arguments to be passed to the ``httpx.Client`` and ``httpx.AsyncClient`` constructor.


    Attributes:
        raise_on_unexpected_status: Whether or not to raise an errors.UnexpectedStatus if the API returns a
            status code that was not documented in the source OpenAPI document. Can also be provided as a keyword
            argument to the constructor.
        token: The token to use for authentication
        prefix: The prefix to use for the Authorization header
        auth_header_name: The name of the Authorization header
    """

    raise_on_unexpected_status: bool = field(default=False, kw_only=True)
    _base_url: str = field(alias="base_url")
    _cookies: dict[str, str] = field(factory=dict, kw_only=True, alias="cookies")
    _headers: dict[str, str] = field(factory=dict, kw_only=True, alias="headers")
    _timeout: Optional[httpx.Timeout] = field(
        default=None, kw_only=True, alias="timeout"
    )
    _verify_ssl: Union[str, bool, ssl.SSLContext] = field(
        default=True, kw_only=True, alias="verify_ssl"
    )
    _follow_redirects: bool = field(
        default=False, kw_only=True, alias="follow_redirects"
    )
    _httpx_args: dict[str, Any] = field(factory=dict, kw_only=True, alias="httpx_args")
    _client: Optional[httpx.Client] = field(default=None, init=False)
    _async_client: Optional[httpx.AsyncClient] = field(default=None, init=False)

    token: str
    prefix: str = "Bearer"
    auth_header_name: str = "Authorization"

    def with_headers(self, headers: dict[str, str]) -> "AuthenticatedClient":
        """Get a new client matching this one with additional headers"""
        if self._client is not None:
            self._client.headers.update(headers)
        if self._async_client is not None:
            self._async_client.headers.update(headers)
        return evolve(self, headers={**self._headers, **headers})

    def with_cookies(self, cookies: dict[str, str]) -> "AuthenticatedClient":
        """Get a new client matching this one with additional cookies"""
        if self._client is not None:
            self._client.cookies.update(cookies)
        if self._async_client is not None:
            self._async_client.cookies.update(cookies)
        return evolve(self, cookies={**self._cookies, **cookies})

    def with_timeout(self, timeout: httpx.Timeout) -> "AuthenticatedClient":
        """Get a new client matching this one with a new timeout (in seconds)"""
        if self._client is not None:
            self._client.timeout = timeout
        if self._async_client is not None:
            self._async_client.timeout = timeout
        return evolve(self, timeout=timeout)

    def set_httpx_client(self, client: httpx.Client) -> "AuthenticatedClient":
        """Manually set the underlying httpx.Client

        **NOTE**: This will override any other settings on the client, including cookies, headers, and timeout.
        """
        self._client = client
        return self

    def _update_user_agent(self) -> None:
        """Update the User-Agent header to include the SDK identifier if not present."""
        sdk_prefix = "python-sdk/"
        sdk_identifier = f"{sdk_prefix}{VERSION}"
        if "User-Agent" not in self._headers:
            self._headers["User-Agent"] = sdk_identifier
        elif sdk_prefix not in self._headers["User-Agent"]:
            self._headers["User-Agent"] = (
                f"{self._headers['User-Agent']} {sdk_identifier}"
            )

    def get_httpx_client(self) -> httpx.Client:
        """Get the underlying httpx.Client, constructing a new one if not previously set"""
        if self._client is None:
            self._headers[self.auth_header_name] = (
                f"{self.prefix} {self.token}" if self.prefix else self.token
            )
            self._update_user_agent()
            self._client = httpx.Client(
                base_url=self._base_url,
                cookies=self._cookies,
                headers=self._headers,
                timeout=self._timeout,
                verify=self._verify_ssl,
                follow_redirects=self._follow_redirects,
                **self._httpx_args,
            )
        return self._client

    def __enter__(self) -> "AuthenticatedClient":
        """Enter a context manager for self.client—you cannot enter twice (see httpx docs)"""
        self.get_httpx_client().__enter__()
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        """Exit a context manager for internal httpx.Client (see httpx docs)"""
        self.get_httpx_client().__exit__(*args, **kwargs)

    def set_async_httpx_client(
        self, async_client: httpx.AsyncClient
    ) -> "AuthenticatedClient":
        """Manually the underlying httpx.AsyncClient

        **NOTE**: This will override any other settings on the client, including cookies, headers, and timeout.
        """
        self._async_client = async_client
        return self

    def get_async_httpx_client(self) -> httpx.AsyncClient:
        """Get the underlying httpx.AsyncClient, constructing a new one if not previously set"""
        if self._async_client is None:
            self._headers[self.auth_header_name] = (
                f"{self.prefix} {self.token}" if self.prefix else self.token
            )
            self._update_user_agent()
            self._async_client = httpx.AsyncClient(
                base_url=self._base_url,
                cookies=self._cookies,
                headers=self._headers,
                timeout=self._timeout,
                verify=self._verify_ssl,
                follow_redirects=self._follow_redirects,
                **self._httpx_args,
            )
        return self._async_client

    async def __aenter__(self) -> "AuthenticatedClient":
        """Enter a context manager for underlying httpx.AsyncClient—you cannot enter twice (see httpx docs)"""
        await self.get_async_httpx_client().__aenter__()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        """Exit a context manager for underlying httpx.AsyncClient (see httpx docs)"""
        await self.get_async_httpx_client().__aexit__(*args, **kwargs)


class VizQLDataServiceClient:
    """Client for VizQL Data Service API."""

    def __init__(
        self,
        url: str,
        server: TSC.Server,
        auth: Union[TSC.JWTAuth, TSC.PersonalAccessTokenAuth, TSC.TableauAuth],
        verify_ssl: Union[str, bool, ssl.SSLContext] = True,
    ):
        """Initialize the client.

        Args:
            url: The base URL of the server.
            server: The Tableau Server instance.
            auth: The authentication object. Can be one of:
                - TSC.JWTAuth: JWT authentication (--jwt-token)
                - TSC.PersonalAccessTokenAuth: Personal access token authentication (--pat-name, --pat-secret)
                - TSC.TableauAuth: Tableau authentication (--user, --password)
            verify_ssl: Whether or not to verify the SSL certificate of the API server.
                This should be True in production, but can be set to False for testing purposes.
                Can also be a path to a CA bundle file or an ssl.SSLContext instance.
        """
        self.url = url
        self.server = server
        self.auth = auth
        self.verify_ssl = verify_ssl
        self._client = self._create_client()
        self.raise_on_unexpected_status = False

    def _create_client(self) -> AuthenticatedClient:
        """Create an authenticated client with proper server URL."""
        base_url = format_server_url(self.url, API_SUBDOMAIN)
        print(f"Server URL: {base_url}")

        return AuthenticatedClient(
            base_url=base_url,
            token=self.server.auth_token,
            prefix="",
            auth_header_name=X_TABLEAU_AUTH,
            verify_ssl=self.verify_ssl,
        )

    @property
    def client(self) -> AuthenticatedClient:
        """Get the authenticated client.

        Returns:
            AuthenticatedClient: The authenticated client instance.

        Raises:
            RuntimeError: If not signed in.
        """
        return self._client

    def get_httpx_client(self) -> httpx.Client:
        """Get the underlying httpx.Client."""
        return self.client.get_httpx_client()

    def get_async_httpx_client(self) -> httpx.AsyncClient:
        """Get the underlying httpx.AsyncClient."""
        return self.client.get_async_httpx_client()

    def __str__(self) -> str:
        return f"Client for {self.auth}"



# generated by datamodel-codegen:
#   filename:  VizQLDataServiceOpenAPISchema.json
#   timestamp: 2025-08-20T20:30:50+00:00

from __future__ import annotations

from datetime import date
from enum import Enum
from typing import Annotated, Any, Dict, List, Optional, Union

from .tableau_model import TableauModel
from pydantic import AwareDatetime, Field as PydanticField, RootModel


class TableauError(TableauModel):
    errorCode: Optional[str] = None
    message: Optional[str] = None
    datetime: Optional[AwareDatetime] = None
    debug: Optional[Dict[str, Any]] = None
    tab_error_code: Annotated[Optional[str], PydanticField(alias='tab-error-code')] = None


class DataType(Enum):
    INTEGER = 'INTEGER'
    REAL = 'REAL'
    STRING = 'STRING'
    DATETIME = 'DATETIME'
    BOOLEAN = 'BOOLEAN'
    DATE = 'DATE'
    SPATIAL = 'SPATIAL'
    UNKNOWN = 'UNKNOWN'


class Connection(TableauModel):
    connectionLuid: Optional[str] = None
    connectionUsername: str
    connectionPassword: str


class Datasource(TableauModel):
    datasourceLuid: Annotated[
        str, PydanticField(description='The LUID of the data source to be queried.')
    ]
    connections: Optional[List[Connection]] = None


class FilterType(Enum):
    QUANTITATIVE_DATE = 'QUANTITATIVE_DATE'
    QUANTITATIVE_NUMERICAL = 'QUANTITATIVE_NUMERICAL'
    SET = 'SET'
    MATCH = 'MATCH'
    DATE = 'DATE'
    TOP = 'TOP'


class DimensionFilterField(TableauModel):
    fieldCaption: Annotated[
        str, PydanticField(description='The caption of the field to filter on.')
    ]


class CalculatedFilterField(TableauModel):
    calculation: Annotated[
        str, PydanticField(description='A Tableau calculation that will be used to filter on.')
    ]


class Function(Enum):
    SUM = 'SUM'
    AVG = 'AVG'
    MEDIAN = 'MEDIAN'
    COUNT = 'COUNT'
    COUNTD = 'COUNTD'
    MIN = 'MIN'
    MAX = 'MAX'
    STDEV = 'STDEV'
    VAR = 'VAR'
    COLLECT = 'COLLECT'
    YEAR = 'YEAR'
    QUARTER = 'QUARTER'
    MONTH = 'MONTH'
    WEEK = 'WEEK'
    DAY = 'DAY'
    TRUNC_YEAR = 'TRUNC_YEAR'
    TRUNC_QUARTER = 'TRUNC_QUARTER'
    TRUNC_MONTH = 'TRUNC_MONTH'
    TRUNC_WEEK = 'TRUNC_WEEK'
    TRUNC_DAY = 'TRUNC_DAY'
    AGG = 'AGG'
    NONE = 'NONE'
    UNSPECIFIED = 'UNSPECIFIED'


class QuantitativeFilterType(Enum):
    RANGE = 'RANGE'
    MIN = 'MIN'
    MAX = 'MAX'
    ONLY_NULL = 'ONLY_NULL'
    ONLY_NON_NULL = 'ONLY_NON_NULL'


class QueryOutput(TableauModel):
    data: Optional[List] = None


class ReturnFormat(Enum):
    OBJECTS = 'OBJECTS'
    ARRAYS = 'ARRAYS'


class SortDirection(Enum):
    ASC = 'ASC'
    DESC = 'DESC'


class PeriodType(Enum):
    MINUTES = 'MINUTES'
    HOURS = 'HOURS'
    DAYS = 'DAYS'
    WEEKS = 'WEEKS'
    MONTHS = 'MONTHS'
    QUARTERS = 'QUARTERS'
    YEARS = 'YEARS'


class DateRangeType(Enum):
    CURRENT = 'CURRENT'
    LAST = 'LAST'
    LASTN = 'LASTN'
    NEXT = 'NEXT'
    NEXTN = 'NEXTN'
    TODATE = 'TODATE'


class Direction(Enum):
    TOP = 'TOP'
    BOTTOM = 'BOTTOM'


class FieldBase(TableauModel):
    fieldCaption: Annotated[
        str,
        PydanticField(
            description='Either the name of a specific Field in the data source, or, in the case of a calculation, a user-supplied name for the calculation.'
        ),
    ]
    fieldAlias: Annotated[
        Optional[str],
        PydanticField(
            description='An alternative name to give the field. Will only be used in object format output.'
        ),
    ] = None
    maxDecimalPlaces: Annotated[
        Optional[int],
        PydanticField(
            description='The maximum number of decimal places. Any trailing 0s will be dropped. The maxDecimalPlaces value must be greater or equal to 0.'
        ),
    ] = None
    sortDirection: Optional[SortDirection] = None
    sortPriority: Annotated[
        Optional[int],
        PydanticField(
            description='To enable sorting on a specific Field, provide a sortPriority for that field, and that field will be sorted. The sortPriority provides a ranking of how to sort fields when multiple fields are being sorted. The highest priority (lowest number) field is sorted first. If only one field is being sorted, then any value may be used for sortPriority. SortPriority should be an integer greater than 0.'
        ),
    ] = None


class DimensionField(FieldBase):
    fieldCaption: Optional[Any] = None
    fieldAlias: Optional[Any] = None
    maxDecimalPlaces: Optional[Any] = None
    sortDirection: Optional[Any] = None
    sortPriority: Optional[Any] = None


class MeasureField(FieldBase):
    function: Function
    fieldCaption: Optional[Any] = None
    fieldAlias: Optional[Any] = None
    maxDecimalPlaces: Optional[Any] = None
    sortDirection: Optional[Any] = None
    sortPriority: Optional[Any] = None


class CalculatedField(FieldBase):
    calculation: Annotated[
        str,
        PydanticField(
            description='A Tableau calculation that will be returned as a Field in the query.'
        ),
    ]
    fieldCaption: Optional[Any] = None
    fieldAlias: Optional[Any] = None
    maxDecimalPlaces: Optional[Any] = None
    sortDirection: Optional[Any] = None
    sortPriority: Optional[Any] = None


class FieldMetadata(TableauModel):
    fieldName: Optional[str] = None
    fieldCaption: Optional[str] = None
    dataType: Optional[DataType] = None
    defaultAggregation: Optional[Function] = None
    logicalTableId: Optional[str] = None


class MeasureFilterField(TableauModel):
    fieldCaption: Annotated[
        str, PydanticField(description='The caption of the field to filter on.')
    ]
    function: Function


class MetadataOutput(TableauModel):
    data: Optional[List[FieldMetadata]] = None


class QueryOptions(TableauModel):
    returnFormat: Optional[ReturnFormat] = None
    debug: Optional[bool] = False


class ReadMetadataRequest(TableauModel):
    datasource: Datasource
    options: Optional[QueryOptions] = None


class Field(RootModel[Union[DimensionField, MeasureField, CalculatedField]]):
    root: Union[DimensionField, MeasureField, CalculatedField]


class FilterField(
    RootModel[Union[DimensionFilterField, MeasureFilterField, CalculatedFilterField]]
):
    root: Union[DimensionFilterField, MeasureFilterField, CalculatedFilterField]


class QueryDatasourceOptions(QueryOptions):
    disaggregate: Optional[bool] = False


class Filter(TableauModel):
    field: FilterField
    filterType: FilterType
    context: Annotated[
        Optional[bool],
        PydanticField(
            description="Make the given filter a context filter, meaning that it's an independent filter. Any other filters that you set are defined as dependent filters because they process only the data that passes through the context filter."
        ),
    ] = False


class MatchFilter(Filter):
    contains: Annotated[
        Optional[str], PydanticField(description='Matches when a field contains this value.')
    ] = None
    startsWith: Annotated[
        Optional[str], PydanticField(description='Matches when a field starts with this value.')
    ] = None
    endsWith: Annotated[
        Optional[str], PydanticField(description='Matches when a field ends with this value.')
    ] = None
    exclude: Annotated[
        Optional[bool],
        PydanticField(description='When true, the inverse of the matching logic will be used.'),
    ] = False


class QuantitativeFilterBase(Filter):
    quantitativeFilterType: QuantitativeFilterType
    includeNulls: Annotated[
        Optional[bool],
        PydanticField(
            description='Should nulls be returned or not. Only applies to RANGE, MIN, and MAX filters. If not provided, the default is to not include null values.'
        ),
    ] = None


class QuantitativeNumericalFilter(QuantitativeFilterBase):
    min: Annotated[
        Optional[float],
        PydanticField(
            description='A numerical value, either integer or floating point, indicating the minimum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN.'
        ),
    ] = None
    max: Annotated[
        Optional[float],
        PydanticField(
            description='A numerical value, either integer or floating point, indicating the maximum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN.'
        ),
    ] = None


class QuantitativeDateFilter(QuantitativeFilterBase):
    minDate: Annotated[
        Optional[date],
        PydanticField(
            description='An RFC 3339 date indicating the earliest date to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN.'
        ),
    ] = None
    maxDate: Annotated[
        Optional[date],
        PydanticField(
            description='An RFC 3339 date indicating the latest date to filter on. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN.'
        ),
    ] = None


class Query(TableauModel):
    fields: Annotated[
        List[Field], PydanticField(description='An array of fields that define the query.')
    ]
    filters: Annotated[
        Optional[List[TabFilter]],
        PydanticField(description='An optional array of filters to apply to the query.'),
    ] = None


class QueryRequest(TableauModel):
    datasource: Datasource
    query: Query
    options: Optional[QueryDatasourceOptions] = None


class SetFilter(Filter):
    values: Annotated[List[Any], PydanticField(description='An array of values to filter on.')]
    exclude: Optional[bool] = False


class RelativeDateFilter(Filter):
    periodType: Annotated[
        PeriodType, PydanticField(description='The units of time in the relative date range.')
    ]
    dateRangeType: Annotated[
        DateRangeType, PydanticField(description='The direction in the relative date range.')
    ]
    rangeN: Annotated[
        Optional[int],
        PydanticField(
            description='When dateRangeType is LASTN or NEXTN, this is the N value (how many years, months, etc.).'
        ),
    ] = None
    anchorDate: Annotated[
        Optional[date],
        PydanticField(
            description="If a value for this field isn't provided, the value defaults to today."
        ),
    ] = None
    includeNulls: Annotated[
        Optional[bool],
        PydanticField(
            description="Should nulls be returned or not. If a value isn't provided, the default is to not include null values."
        ),
    ] = None


class TopNFilter(Filter):
    direction: Annotated[
        Optional[Direction],
        PydanticField(description='Top (ascending) or Bottom (descending) N.'),
    ] = 'TOP'
    howMany: Annotated[
        int,
        PydanticField(
            description='The number of values from the top or the bottom of the given fieldToMeasure.'
        ),
    ]
    fieldToMeasure: FilterField

class TabFilter(RootModel[Union[
    MatchFilter, QuantitativeNumericalFilter, QuantitativeDateFilter, SetFilter, RelativeDateFilter, TopNFilter]]):
    root: Union[
        MatchFilter, QuantitativeNumericalFilter, QuantitativeDateFilter, SetFilter, RelativeDateFilter, TopNFilter]



import os
import sys

import tableauserverclient as TSC

# Add project root to sys.path
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, root_dir)

# Determine if we're in development or production environment
is_development = os.path.basename(root_dir) == "python_sdk"

if is_development:
    import src.examples.common as common
    from src.api import query_datasource, read_metadata
    from src.api.client import VizQLDataServiceClient
    from src.api.openapi_generated import QueryRequest, ReadMetadataRequest
    from src.examples.payload import QUERY_FUNCTIONS
else:
    import vizql_data_service_py.examples.common as common  # type: ignore
    from vizql_data_service_py.api import query_datasource, read_metadata  # type: ignore
    from vizql_data_service_py.api.client import VizQLDataServiceClient  # type: ignore
    from vizql_data_service_py.api.openapi_generated import QueryRequest, ReadMetadataRequest  # type: ignore
    from vizql_data_service_py.examples.payload import QUERY_FUNCTIONS  # type: ignore


async def execute(args):
    server_url, auth = common.create_server(
        url=args.server,
        username=args.user,
        password=args.password,
        pat_name=args.pat_name,
        pat_secret=args.pat_secret,
        jwt_token=args.jwt_token,
        site_id=args.site,
    )

    server = TSC.Server(server_url)

    with server.auth.sign_in(auth):
        client = VizQLDataServiceClient(server_url, server, auth)
        datasource_luid = common.list_datasources_and_get_luid(server, args.verbose)
        datasource = common.create_datasource(datasource_luid)

        # Read metadata example
        try:
            print("\n=== ReadMetadata Query ===")
            metadata_request = ReadMetadataRequest(datasource=datasource)
            print(f"Request Body: {metadata_request}")

            metadata_response = await read_metadata.asyncio_detailed(
                client=client, body=metadata_request
            )
            common.handle_response(
                metadata_response, "ReadMetadata Query", args.verbose
            )
        except Exception as e:
            common.handle_error(e, "ReadMetadata Query", args.verbose)

        # Query data source examples
        async def execute_query(query_func):
            try:
                query_request = QueryRequest(query=query_func(), datasource=datasource)
                print(f"\n=== ExecuteQuery: {query_func.__name__} ===")
                if args.verbose:
                    print(f"Request Body: {query_request}")

                response = await query_datasource.asyncio_detailed(
                    client=client, body=query_request
                )
                common.handle_response(
                    response, f"Query {query_func.__name__}", args.verbose
                )
            except Exception as e:
                print(f"\n=== ExecuteQuery: {query_func.__name__} ===")
                common.handle_error(e, f"Query {query_func.__name__}", args.verbose)

        # Execute queries sequentially
        for query_func in QUERY_FUNCTIONS:
            await execute_query(query_func)



import asyncio
import os
import sys

# Add project root to sys.path
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, root_dir)

# Determine if we're in development or production environment
is_development = os.path.basename(root_dir) == "python_sdk"

if is_development:
    import src.examples.async_examples as async_examples
    import src.examples.common as common
    import src.examples.sync_examples as sync_examples
else:
    import vizql_data_service_py.examples.async_examples as async_examples  # type: ignore
    import vizql_data_service_py.examples.common as common  # type: ignore
    import vizql_data_service_py.examples.sync_examples as sync_examples  # type: ignore

if __name__ == "__main__":
    # Check for help flag before parsing arguments
    if "-h" in sys.argv or "--help" in sys.argv:
        common.print_help()
        sys.exit(0)

    parser = common.parse_arguments()
    parser.add_argument(
        "--async",
        dest="async_mode",
        action="store_true",
        help="Run examples in async mode",
    )
    args = parser.parse_args()

    if args.async_mode:
        if sys.platform == "win32":
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        asyncio.run(async_examples.execute(args))
    else:
        sync_examples.execute(args)
